@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP, GeneralizedNewtypeDeriving, BangPatterns #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Codec.Archive.Tar.Types
@@ -51,14 +52,20 @@ module Codec.Archive.Tar.Types (
   mapEntries,
   mapEntriesNoFail,
   foldEntries,
+  foldlEntries,
   unfoldEntries,
 
+#ifdef TESTS
+  limitToV7FormatCompat
+#endif
   ) where
 
 import Data.Int      (Int64)
 import Data.Monoid   (Monoid(..))
-import qualified Data.ByteString.Lazy as BS
-import Data.ByteString.Lazy (ByteString)
+import qualified Data.ByteString       as BS
+import qualified Data.ByteString.Char8 as BS.Char8
+import qualified Data.ByteString.Lazy  as LBS
+import Control.DeepSeq
 
 import qualified System.FilePath as FilePath.Native
          ( joinPath, splitDirectories, addTrailingPathSeparator )
@@ -70,6 +77,12 @@ import qualified System.FilePath.Windows as FilePath.Windows
 import System.Posix.Types
          ( FileMode )
 
+#ifdef TESTS
+import Test.QuickCheck
+import Control.Applicative ((<$>), pure, (<*>))
+#endif
+
+
 type FileSize  = Int64
 -- | The number of seconds since the UNIX epoch
 type EpochTime = Int64
@@ -84,24 +97,25 @@ data Entry = Entry {
 
     -- | The path of the file or directory within the archive. This is in a
     -- tar-specific form. Use 'entryPath' to get a native 'FilePath'.
-    entryTarPath :: !TarPath,
+    entryTarPath :: {-# UNPACK #-} !TarPath,
 
     -- | The real content of the entry. For 'NormalFile' this includes the
     -- file data. An entry usually contains a 'NormalFile' or a 'Directory'.
     entryContent :: !EntryContent,
 
     -- | File permissions (Unix style file mode).
-    entryPermissions :: !Permissions,
+    entryPermissions :: {-# UNPACK #-} !Permissions,
 
     -- | The user and group to which this file belongs.
-    entryOwnership :: !Ownership,
+    entryOwnership :: {-# UNPACK #-} !Ownership,
 
     -- | The time the file was last modified.
-    entryTime :: !EpochTime,
+    entryTime :: {-# UNPACK #-} !EpochTime,
 
     -- | The tar format the archive is using.
     entryFormat :: !Format
   }
+  deriving (Eq, Show)
 
 -- | Native 'FilePath' of the file or directory within the archive.
 --
@@ -112,15 +126,18 @@ entryPath = fromTarPath . entryTarPath
 --
 -- Portable archives should contain only 'NormalFile' and 'Directory'.
 --
-data EntryContent = NormalFile      ByteString !FileSize
+data EntryContent = NormalFile      LBS.ByteString {-# UNPACK #-} !FileSize
                   | Directory
                   | SymbolicLink    !LinkTarget
                   | HardLink        !LinkTarget
-                  | CharacterDevice !DevMajor !DevMinor
-                  | BlockDevice     !DevMajor !DevMinor
+                  | CharacterDevice {-# UNPACK #-} !DevMajor
+                                    {-# UNPACK #-} !DevMinor
+                  | BlockDevice     {-# UNPACK #-} !DevMajor
+                                    {-# UNPACK #-} !DevMinor
                   | NamedPipe
-                  | OtherEntryType  !TypeCode ByteString !FileSize
-    deriving (Eq, Ord)
+                  | OtherEntryType  {-# UNPACK #-} !TypeCode LBS.ByteString
+                                    {-# UNPACK #-} !FileSize
+    deriving (Eq, Ord, Show)
 
 data Ownership = Ownership {
     -- | The owner user name. Should be set to @\"\"@ if unknown.
@@ -130,12 +147,12 @@ data Ownership = Ownership {
     groupName :: String,
 
     -- | Numeric owner user id. Should be set to @0@ if unknown.
-    ownerId :: !Int,
+    ownerId :: {-# UNPACK #-} !Int,
 
     -- | Numeric owner group id. Should be set to @0@ if unknown.
-    groupId :: !Int
+    groupId :: {-# UNPACK #-} !Int
   }
-    deriving (Eq, Ord)
+    deriving (Eq, Ord, Show)
 
 -- | There have been a number of extensions to the tar file format over the
 -- years. They all share the basic entry fields and put more meta-data in
@@ -158,7 +175,25 @@ data Format =
      -- archives the standard USTAR/POSIX should be used.
      --
    | GnuFormat
-  deriving (Eq, Ord)
+  deriving (Eq, Ord, Show)
+
+instance NFData Entry where
+  rnf (Entry _ c _ _ _ _) = rnf c
+
+instance NFData EntryContent where
+  rnf x = case x of
+      NormalFile       c _  -> rnflbs c
+      OtherEntryType _ c _  -> rnflbs c
+      _                     -> seq x ()
+    where
+#if MIN_VERSION_bytestring(0,10,0)
+      rnflbs = rnf
+#else
+      rnflbs = foldr (\ !_bs r -> r) () . LBS.toChunks
+#endif
+
+instance NFData Ownership where
+  rnf (Ownership o g _ _) = rnf o `seq` rnf g
 
 -- | @rw-r--r--@ for normal files
 ordinaryFilePermissions :: Permissions
@@ -200,9 +235,9 @@ simpleEntry tarpath content = Entry {
 --
 -- > (fileEntry name content) { fileMode = executableFileMode }
 --
-fileEntry :: TarPath -> ByteString -> Entry
+fileEntry :: TarPath -> LBS.ByteString -> Entry
 fileEntry name fileContent =
-  simpleEntry name (NormalFile fileContent (BS.length fileContent))
+  simpleEntry name (NormalFile fileContent (LBS.length fileContent))
 
 -- | A tar 'Entry' for a directory.
 --
@@ -239,10 +274,16 @@ directoryEntry name = simpleEntry name Directory
 --
 -- * The directory separator between the prefix and name is /not/ stored.
 --
-data TarPath = TarPath FilePath -- path name, 100 characters max.
-                       FilePath -- path prefix, 155 characters max.
+data TarPath = TarPath {-# UNPACK #-} !BS.ByteString -- path name, 100 characters max.
+                       {-# UNPACK #-} !BS.ByteString -- path prefix, 155 characters max.
   deriving (Eq, Ord)
 
+instance NFData TarPath where
+  rnf (TarPath _ _) = () -- fully strict by construction
+
+instance Show TarPath where
+  show = show . fromTarPath
+
 -- | Convert a 'TarPath' to a native 'FilePath'.
 --
 -- The native 'FilePath' will use the native directory separator but it is not
@@ -256,15 +297,17 @@ data TarPath = TarPath FilePath -- path name, 100 characters max.
 --   responsibility to check for these conditions (eg using 'checkSecurity').
 --
 fromTarPath :: TarPath -> FilePath
-fromTarPath (TarPath name prefix) = adjustDirectory $
+fromTarPath (TarPath namebs prefixbs) = adjustDirectory $
   FilePath.Native.joinPath $ FilePath.Posix.splitDirectories prefix
                           ++ FilePath.Posix.splitDirectories name
   where
+    name   = BS.Char8.unpack namebs
+    prefix = BS.Char8.unpack prefixbs
     adjustDirectory | FilePath.Posix.hasTrailingPathSeparator name
                     = FilePath.Native.addTrailingPathSeparator
                     | otherwise = id
 
--- | Convert a 'TarPath' to a Unix/Posix 'FilePath'.
+-- | Convert a 'TarPath' to a Unix\/Posix 'FilePath'.
 --
 -- The difference compared to 'fromTarPath' is that it always returns a Unix
 -- style path irrespective of the current operating system.
@@ -273,10 +316,12 @@ fromTarPath (TarPath name prefix) = adjustDirectory $
 -- operating system, eg to perform portability checks.
 --
 fromTarPathToPosixPath :: TarPath -> FilePath
-fromTarPathToPosixPath (TarPath name prefix) = adjustDirectory $
+fromTarPathToPosixPath (TarPath namebs prefixbs) = adjustDirectory $
   FilePath.Posix.joinPath $ FilePath.Posix.splitDirectories prefix
                          ++ FilePath.Posix.splitDirectories name
   where
+    name   = BS.Char8.unpack namebs
+    prefix = BS.Char8.unpack prefixbs
     adjustDirectory | FilePath.Posix.hasTrailingPathSeparator name
                     = FilePath.Posix.addTrailingPathSeparator
                     | otherwise = id
@@ -290,10 +335,12 @@ fromTarPathToPosixPath (TarPath name prefix) = adjustDirectory $
 -- operating system, eg to perform portability checks.
 --
 fromTarPathToWindowsPath :: TarPath -> FilePath
-fromTarPathToWindowsPath (TarPath name prefix) = adjustDirectory $
+fromTarPathToWindowsPath (TarPath namebs prefixbs) = adjustDirectory $
   FilePath.Windows.joinPath $ FilePath.Posix.splitDirectories prefix
                            ++ FilePath.Posix.splitDirectories name
   where
+    name   = BS.Char8.unpack namebs
+    prefix = BS.Char8.unpack prefixbs
     adjustDirectory | FilePath.Posix.hasTrailingPathSeparator name
                     = FilePath.Windows.addTrailingPathSeparator
                     | otherwise = id
@@ -325,11 +372,13 @@ splitLongPath :: FilePath -> Either String TarPath
 splitLongPath path =
   case packName nameMax (reverse (FilePath.Posix.splitPath path)) of
     Left err                 -> Left err
-    Right (name, [])         -> Right (TarPath name "")
+    Right (name, [])         -> Right $! TarPath (BS.Char8.pack name)
+                                                  BS.empty
     Right (name, first:rest) -> case packName prefixMax remainder of
       Left err               -> Left err
       Right (_     , (_:_))  -> Left "File name too long (cannot split)"
-      Right (prefix, [])     -> Right (TarPath name prefix)
+      Right (prefix, [])     -> Right $! TarPath (BS.Char8.pack name)
+                                                 (BS.Char8.pack prefix)
       where
         -- drop the '/' between the name and prefix:
         remainder = init first : rest
@@ -353,23 +402,31 @@ splitLongPath path =
 -- | The tar format allows just 100 ASCII characters for the 'SymbolicLink' and
 -- 'HardLink' entry types.
 --
-newtype LinkTarget = LinkTarget FilePath
-  deriving (Eq, Ord)
+newtype LinkTarget = LinkTarget BS.ByteString
+  deriving (Eq, Ord, Show)
+
+instance NFData LinkTarget where
+#if MIN_VERSION_bytestring(0,10,0)
+    rnf (LinkTarget bs) = rnf bs
+#else
+    rnf (LinkTarget !_bs) = ()
+#endif
 
 -- | Convert a native 'FilePath' to a tar 'LinkTarget'. This may fail if the
 -- string is longer than 100 characters or if it contains non-portable
 -- characters.
 --
 toLinkTarget   :: FilePath -> Maybe LinkTarget
-toLinkTarget path | length path <= 100 = Just (LinkTarget path)
+toLinkTarget path | length path <= 100 = Just $! LinkTarget (BS.Char8.pack path)
                   | otherwise          = Nothing
 
 -- | Convert a tar 'LinkTarget' to a native 'FilePath'.
 --
 fromLinkTarget :: LinkTarget -> FilePath
-fromLinkTarget (LinkTarget path) = adjustDirectory $
+fromLinkTarget (LinkTarget pathbs) = adjustDirectory $
   FilePath.Native.joinPath $ FilePath.Posix.splitDirectories path
   where
+    path = BS.Char8.unpack pathbs
     adjustDirectory | FilePath.Posix.hasTrailingPathSeparator path
                     = FilePath.Native.addTrailingPathSeparator
                     | otherwise = id
@@ -377,9 +434,10 @@ fromLinkTarget (LinkTarget path) = adjustDirectory $
 -- | Convert a tar 'LinkTarget' to a Unix/Posix 'FilePath'.
 --
 fromLinkTargetToPosixPath :: LinkTarget -> FilePath
-fromLinkTargetToPosixPath (LinkTarget path) = adjustDirectory $
+fromLinkTargetToPosixPath (LinkTarget pathbs) = adjustDirectory $
   FilePath.Posix.joinPath $ FilePath.Posix.splitDirectories path
   where
+    path = BS.Char8.unpack pathbs
     adjustDirectory | FilePath.Posix.hasTrailingPathSeparator path
                     = FilePath.Native.addTrailingPathSeparator
                     | otherwise = id
@@ -387,9 +445,10 @@ fromLinkTargetToPosixPath (LinkTarget path) = adjustDirectory $
 -- | Convert a tar 'LinkTarget' to a Windows 'FilePath'.
 --
 fromLinkTargetToWindowsPath :: LinkTarget -> FilePath
-fromLinkTargetToWindowsPath (LinkTarget path) = adjustDirectory $
+fromLinkTargetToWindowsPath (LinkTarget pathbs) = adjustDirectory $
   FilePath.Windows.joinPath $ FilePath.Posix.splitDirectories path
   where
+    path = BS.Char8.unpack pathbs
     adjustDirectory | FilePath.Posix.hasTrailingPathSeparator path
                     = FilePath.Windows.addTrailingPathSeparator
                     | otherwise = id
@@ -417,6 +476,7 @@ fromLinkTargetToWindowsPath (LinkTarget path) = adjustDirectory $
 data Entries e = Next Entry (Entries e)
                | Done
                | Fail e
+  deriving (Eq, Show)
 
 infixr 5 `Next`
 
@@ -425,7 +485,7 @@ infixr 5 `Next`
 -- return.
 --
 -- It can be used to generate 'Entries' from some other type. For example it is
--- used internally to lazily unfold entries from a 'ByteString'.
+-- used internally to lazily unfold entries from a 'LBS.ByteString'.
 --
 unfoldEntries :: (a -> Either e (Maybe (Entry, a))) -> a -> Entries e
 unfoldEntries f = unfold
@@ -449,6 +509,17 @@ foldEntries next done fail' = fold
     fold Done        = done
     fold (Fail err)  = fail' err
 
+-- | A 'foldl'-like function on Entries. It either returns the final
+-- accumulator result, or the failure along with the intermediate accumulator
+-- value.
+--
+foldlEntries :: (a -> Entry -> a) -> a -> Entries e -> Either (e, a) a
+foldlEntries f z = go z
+  where
+    go !acc (Next e es) = go (f acc e) es
+    go !acc  Done       = Right acc
+    go !acc (Fail err)  = Left (err, acc)
+
 -- | This is like the standard 'map' function on lists, but for 'Entries'. It
 -- includes failure as a extra possible outcome of the mapping function.
 --
@@ -471,3 +542,156 @@ instance Monoid (Entries e) where
 instance Functor Entries where
   fmap f = foldEntries Next Done (Fail . f)
 
+instance NFData e => NFData (Entries e) where
+  rnf (Next e es) = rnf e `seq` rnf es
+  rnf  Done       = ()
+  rnf (Fail e)    = rnf e
+
+
+-------------------------
+-- QuickCheck instances
+--
+
+#ifdef TESTS
+
+instance Arbitrary Entry where
+  arbitrary = Entry <$> arbitrary <*> arbitrary <*> arbitraryPermissions
+                    <*> arbitrary <*> arbitraryEpochTime <*> arbitrary
+    where
+      arbitraryPermissions :: Gen Permissions
+      arbitraryPermissions = fromIntegral <$> (arbitraryOctal 7 :: Gen Int)
+
+      arbitraryEpochTime :: Gen EpochTime
+      arbitraryEpochTime = fromIntegral <$> (arbitraryOctal 11 :: Gen Int)
+
+  shrink (Entry path content perms author time format) =
+      [ Entry path' content' perms author' time' format 
+      | (path', content', author', time') <-
+         shrink (path, content, author, time) ]
+   ++ [ Entry path content perms' author time format
+      | perms' <- shrinkIntegral perms ]
+
+instance Arbitrary TarPath where
+  arbitrary = either error id
+            . toTarPath False
+            . FilePath.Posix.joinPath
+          <$> listOf1ToN (255 `div` 5)
+                         (elements (map (replicate 4) "abcd"))
+
+  shrink = map (either error id . toTarPath False)
+         . map FilePath.Posix.joinPath
+         . filter (not . null)
+         . shrinkList shrinkNothing
+         . FilePath.Posix.splitPath
+         . fromTarPathToPosixPath
+
+instance Arbitrary LinkTarget where
+  arbitrary = maybe (error "link target too large") id
+            . toLinkTarget
+            . FilePath.Native.joinPath
+          <$> listOf1ToN (100 `div` 5)
+                         (elements (map (replicate 4) "abcd"))
+
+  shrink = map (maybe (error "link target too large") id . toLinkTarget)
+         . map FilePath.Posix.joinPath
+         . filter (not . null)
+         . shrinkList shrinkNothing
+         . FilePath.Posix.splitPath
+         . fromLinkTargetToPosixPath
+
+
+listOf1ToN :: Int -> Gen a -> Gen [a]
+listOf1ToN n g = sized $ \sz -> do
+    n <- choose (1, min n (max 1 sz))
+    vectorOf n g
+
+listOf0ToN :: Int -> Gen a -> Gen [a]
+listOf0ToN n g = sized $ \sz -> do
+    n <- choose (0, min n sz)
+    vectorOf n g
+
+instance Arbitrary EntryContent where
+  arbitrary =
+    frequency
+      [ (16, do bs <- arbitrary;
+                return (NormalFile bs (LBS.length bs)))
+      , (2, pure Directory)
+      , (1, SymbolicLink    <$> arbitrary)
+      , (1, HardLink        <$> arbitrary)
+      , (1, CharacterDevice <$> arbitraryOctal 7 <*> arbitraryOctal 7)
+      , (1, BlockDevice     <$> arbitraryOctal 7 <*> arbitraryOctal 7)
+      , (1, pure NamedPipe)
+      , (1, do c  <- elements (['A'..'Z']++['a'..'z'])
+               bs <- arbitrary;
+               return (OtherEntryType c bs (LBS.length bs)))
+      ]
+
+  shrink (NormalFile bs _)   = [ NormalFile bs' (LBS.length bs') 
+                               | bs' <- shrink bs ]
+  shrink  Directory          = []
+  shrink (SymbolicLink link) = [ SymbolicLink link' | link' <- shrink link ]
+  shrink (HardLink     link) = [ HardLink     link' | link' <- shrink link ]
+  shrink (CharacterDevice ma mi) = [ CharacterDevice ma' mi'
+                                   | (ma', mi') <- shrink (ma, mi) ]
+  shrink (BlockDevice     ma mi) = [ BlockDevice ma' mi'
+                                   | (ma', mi') <- shrink (ma, mi) ]
+  shrink  NamedPipe              = []
+  shrink (OtherEntryType c bs _) = [ OtherEntryType c bs' (LBS.length bs') 
+                                   | bs' <- shrink bs ]
+
+instance Arbitrary LBS.ByteString where
+  arbitrary = fmap LBS.pack arbitrary
+  shrink    = map LBS.pack . shrink . LBS.unpack
+
+instance Arbitrary BS.ByteString where
+  arbitrary = fmap BS.pack arbitrary
+  shrink    = map BS.pack . shrink . BS.unpack
+
+instance Arbitrary Ownership where
+  arbitrary = Ownership <$> name <*> name
+                        <*> idno <*> idno
+    where
+      name = listOf0ToN 32 (arbitrary `suchThat` (/= '\0'))
+      idno = arbitraryOctal 7
+
+  shrink (Ownership oname gname oid gid) =
+    [ Ownership oname' gname' oid' gid'
+    | (oname', gname', oid', gid') <- shrink (oname, gname, oid, gid) ]
+
+instance Arbitrary Format where
+  arbitrary = elements [V7Format, UstarFormat, GnuFormat]
+
+
+--arbitraryOctal :: (Integral n, Random n) => Int -> Gen n
+arbitraryOctal n =
+    oneof [ pure 0
+          , choose (0, upperBound)
+          , pure upperBound
+          ]
+  where
+    upperBound = 8^n-1
+
+-- For QC tests it's useful to have a way to limit the info to that which can
+-- be expressed in the old V7 format
+limitToV7FormatCompat :: Entry -> Entry
+limitToV7FormatCompat entry@Entry { entryFormat = V7Format } =
+    entry {
+      entryContent = case entryContent entry of
+        CharacterDevice _ _ -> OtherEntryType  '3' LBS.empty 0
+        BlockDevice     _ _ -> OtherEntryType  '4' LBS.empty 0
+        Directory           -> OtherEntryType  '5' LBS.empty 0
+        NamedPipe           -> OtherEntryType  '6' LBS.empty 0
+        other               -> other,
+
+      entryOwnership = (entryOwnership entry) {
+        groupName = "",
+        ownerName = ""
+      },
+
+      entryTarPath = let TarPath name _prefix = entryTarPath entry
+                      in TarPath name BS.empty
+    }
+limitToV7FormatCompat entry = entry
+
+#endif
+
